import datetime


class hybridlogger:

    @staticmethod
    def ha_log(logger, hass_api, level, message):
        if hass_api and hasattr(hass_api, 'log'):
            hass_api.log(message, level=level)
        else:
            if level == "DEBUG":
                logger.debug(f"@ {datetime.datetime.now().isoformat()} {message}")
            elif level == "INFO":
                logger.info(f"I {datetime.datetime.now().isoformat()} {message}")
            elif level == "WARNING":
                logger.warning(f"W {datetime.datetime.now().isoformat()} {message}")
            elif level == "ERROR":
                logger.error(f"E {datetime.datetime.now().isoformat()} {message}")





class LocalProxyBasePlugin(type):
    def __init__(cls, name, bases, attrs):
        super(LocalProxyBasePlugin, cls).__init__(name)
        if not hasattr(cls, 'localproxy_plugins'):
            cls.localproxy_plugins = []
        else:
            cls.register(cls)  # Called when a plugin class is imported

    def register(cls, plugin):
        cls.localproxy_plugins.append(plugin())


class LocalProxyPlugin(object, metaclass=LocalProxyBasePlugin):

    config = None
    logger = None
    hass_api = None
    semaphore = None
    client = None

    def terminate(self):
        pass



from dsmr_parser.parsers import TelegramParser
from dsmr_parser.exceptions import ParseError, InvalidChecksumError
from dsmr_parser.clients.telegram_buffer import TelegramBuffer
from dsmr_parser import telegram_specifications
from dsmr_parser.clients.settings import SERIAL_SETTINGS_V2_2, \
    SERIAL_SETTINGS_V4, SERIAL_SETTINGS_V5

import serial

import threading
from omnik.ha_logger import hybridlogger
import time
import socket


class Terminal(object):

    def __init__(self, config, logger, hass_api, terminal_name, dsmr_serial_callback, dsmr_version):
        self.config = config
        self.logger = logger
        self.hass_api = hass_api
        self.terminal_name = terminal_name
        self.dsmr_serial_callback = dsmr_serial_callback
        self.mode = self.config.get(f"dsmr.{self.terminal_name}", 'mode', 'device')
        if self.mode not in ['tcp', 'device']:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "ERROR", f"DSMR terminal {self.terminal_name} mode {self.mode} is not valid. "
                                "Should be tcp or device. Ignoring DSMR configuration!")
            return
        self.device = self.config.get(f"dsmr.{self.terminal_name}", 'device', '/dev/ttyUSB0')
        self.host = self.config.get(f"dsmr.{self.terminal_name}", 'host', 'localhost')
        self.port = self.config.get(f"dsmr.{self.terminal_name}", 'port', '3333')
        self.dsmr_version = dsmr_version
        # start terminal
        self.stop = False
        hybridlogger.ha_log(self.logger, self.hass_api,
                            "INFO", f"Initializing DSMR termimal '{terminal_name}'. Mode: {self.mode}.")
        if self.mode == 'tcp':
            self.thr = threading.Thread(target=self._run_tcp_terminal, name=self.terminal_name)
        elif self.mode == 'device':
            self.thr = threading.Thread(target=self._run_serial_terminal, name=self.terminal_name)
        self.thr.start()

    def terminate(self):
        self.stop = True
        # Wait for Thread to shutdown
        self.thr.join()

    def _get_dsmr_parser(self):
        dsmr_version = self.dsmr_version
        if dsmr_version == '2.2':
            specification = telegram_specifications.V2_2
            serial_settings = SERIAL_SETTINGS_V2_2
        elif dsmr_version == '4':
            specification = telegram_specifications.V4
            serial_settings = SERIAL_SETTINGS_V4
        elif dsmr_version == '5':
            specification = telegram_specifications.V5
            serial_settings = SERIAL_SETTINGS_V5
        elif dsmr_version == '5B':
            specification = telegram_specifications.BELGIUM_FLUVIUS
            serial_settings = SERIAL_SETTINGS_V5
        else:
            raise NotImplementedError("No telegram parser found for version: %s",
                                      dsmr_version)
        self.telegram_parser = TelegramParser(specification)
        serial_settings['timeout'] = 10
        self.serial_settings = serial_settings

    def _dsmr_data_received(self, data):
        """Add incoming data to buffer."""
        data = data.decode('ascii')
        self.telegram_buffer.append(data)

        for telegram in self.telegram_buffer.get_all():
            self._handle_telegram(telegram)

    def _handle_telegram(self, telegram):
        """Send off parsed telegram to handling callback."""
        try:
            parsed_telegram = self.telegram_parser.parse(telegram)
        except InvalidChecksumError as e:
            self.log.warning(str(e))
        except ParseError:
            self.log.exception("failed to parse telegram")
        else:
            self.dsmr_serial_callback(parsed_telegram)

    def _run_tcp_terminal(self):
        self._get_dsmr_parser()
        # buffer to keep incomplete incoming data
        while not self.stop:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "INFO", f"DSMR terminal {self.terminal_name} was started."
                                )
            try:
                self.telegram_buffer = TelegramBuffer()
                server_address = (self.host, self.port)
                self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.sock.settimeout(5)
                self.sock.connect(server_address)
                while not self.stop:
                    data = self.sock.recv(1024)
                    if not data and not self.stop:
                        hybridlogger.ha_log(self.logger, self.hass_api,
                                            "INFO",
                                            f"DSMR terminal {self.terminal_name} was interrupted and will be restarted."
                                            )
                        time.sleep(5)
                        break
                    elif data:
                        self._dsmr_data_received(data)

            except Exception as e:
                if not self.stop:
                    hybridlogger.ha_log(self.logger, self.hass_api,
                                        "WARNING", f"DSMR terminal {self.terminal_name} was interrupted "
                                        f"and will be restarted in a few moments: {e.args}"
                                        )
                    time.sleep(5)
            finally:
                self.sock.close()

    def _run_serial_terminal(self):
        self._get_dsmr_parser()
        # buffer to keep incomplete incoming data
        while not self.stop:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "INFO", f"DSMR terminal {self.terminal_name} was started."
                                )
            try:
                self.telegram_buffer = TelegramBuffer()
                self.sock = serial.Serial(port=self.device, **self.serial_settings)
                while not self.stop:
                    data = self.sock.read_until()
                    if not data and not self.stop:
                        hybridlogger.ha_log(self.logger, self.hass_api,
                                            "INFO",
                                            f"DSMR terminal {self.terminal_name} was interrupted and will be restarted."
                                            )
                        time.sleep(5)
                        break
                    elif data:
                        self._dsmr_data_received(data)

            except Exception as e:
                if not self.stop:
                    hybridlogger.ha_log(self.logger, self.hass_api,
                                        "WARNING", f"DSMR terminal {self.terminal_name} was interrupted "
                                        f"and will be restarted in a few moments: {e.args}"
                                        )
                    time.sleep(5)
            finally:
                self.sock.close()



import os
import sys
import logging
from omnik.ha_logger import hybridlogger
import requests
import json
from .daylight import daylight
import threading
import time
from decimal import Decimal
from datetime import date, datetime, timedelta, timezone

from .plugin_output import Plugin
from .plugin_client import Client
from .dsmr import DSRM
from copy import deepcopy
from cachetools import Cache

import importlib

# from .plugin_localproxy import LocalProxyPlugin

logger = logging.getLogger(__name__)


class DataLogger(object):

    def __init__(self, config, hass_api=None):
        # Defaults to UTC now() - every interval
        self.plant_update = {}
        self.hass_api = hass_api
        self.logger = logger
        self.config = config
        self.cache = Cache(maxsize=10)
        self.start_total_energy = {}
        self.pasttime = datetime.combine(date.today(), datetime.min.time()).timestamp()
        self.dsmr_access = threading.Condition(threading.Lock())

        if self.config.get('default', 'debug', fallback=False):
            logger.setLevel(logging.DEBUG)

        if not self._init_data_fields():
            sys.exit(1)
        # init energy cache
        # total_energy_cache.json
        self.persistant_cache_file = self.config.get('default', 'persistant_cache_file', fallback='./persistant_cache.json')
        self._load_persistant_cache()
        # read data_fields
        try:
            with open(self.data_config_file) as json_file_config:
                self.config.data_field_config = json.load(json_file_config)
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR",
                                f"Error reading configuration file (data_fields.json). Exiting!. Error: {e.args}")
            raise e
            sys.exit(1)

        # read attributes
        if not self._read_attributes():
            sys.exit(1)

        self.every = self._get_interval()
        self.interval_aggregated = self._get_interval_aggregated()

        # Make sure we check for a recent update first
        self.last_update_time = datetime.now(timezone.utc) - timedelta(seconds=self.interval_aggregated)

        # Initialize plugin paths
        sys.path.append(self.__expand_path('plugin_output'))
        sys.path.append(self.__expand_path('plugin_client'))
        sys.path.append(self.__expand_path('plugin_localproxy'))

        self.city = self.config.get('default', 'city', fallback='Amsterdam')
        try:
            self.dl = daylight(self.city)
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", f"City '{self.city}' not recognized. Error: {e}")
            sys.exit(1)

        # Initialize client
        self._init_client()
        # Initialize output plugins
        self._init_output_plugins()

        self.omnik_api_level = 0

        # Init dsmr
        self._init_dsmr()

    def _init_data_fields(self):
        self.config.data_config_file_args = self.config.get('default', 'data_config', fallback='')
        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG",
                            f"Data configuration [args]: '{self.config.data_config_file_args}'.")
        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG",
                            f"Data configuration [path]: '{self.config.data_config_file_path}'.")
        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG",
                            f"Data configuration [shared]: '{self.config.data_config_file_shared}'.")
        if self.config.configfile:
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO", f"Using configuration from '{self.config.configfile}'.")

        if os.path.exists(self.config.data_config_file_args):
            self.data_config_file = self.config.data_config_file_args
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                f"Using configured data configuration from '{self.data_config_file}'.")
        elif os.path.exists(self.config.data_config_file_path):
            self.data_config_file = self.config.data_config_file_path
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                f"Using configured data configuration from '{self.data_config_file}'.")
        elif os.path.exists(self.config.data_config_file_shared):
            self.data_config_file = self.config.data_config_file_shared
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                f"Using shared data configuration from '{self.data_config_file}'.")
        else:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR",
                                "No valid data configuration file (data_fields.json) found. Exiting!")
            return False
        return True

    def _init_attribute_dict(self):
        self.config.attributes = {}
        asset_classes = list(['omnik', 'omnik_dsmr', 'dsmr', 'dsmr_gas'])
        self.config.attributes['asset'] = {
            'omnik': list(['inverter', 'plant_id', 'last_update']),
            'omnik_dsmr': list(['inverter', 'plant_id', 'EQUIPMENT_IDENTIFIER', 'last_update']),
            'dsmr': list(['EQUIPMENT_IDENTIFIER', 'timestamp']),
            'dsmr_gas': list(['EQUIPMENT_IDENTIFIER_GAS', 'timestamp_gas'])
            }
        self.config.attributes['model'] = {
            'omnik': 'Omniksol',
            'omnik_dsmr': 'Omnik data logger',
            'dsmr': 'DSMR electicity meter',
            'dsmr_gas': 'DSMR gas meter'
            }
        self.config.attributes['mf'] = {
            'omnik': 'Omnik',
            'omnik_dsmr': 'JBsoft',
            'dsmr': 'Netbeheer Nederland',
            'dsmr_gas': 'Netbeheer Nederland'
            }
        self.config.attributes['identifier'] = {
            'omnik': 'plant_id',
            'omnik_dsmr': 'plant_id',
            'dsmr': 'EQUIPMENT_IDENTIFIER',
            'dsmr_gas': 'EQUIPMENT_IDENTIFIER_GAS'
            }
        self.config.attributes['devicename'] = {
            'omnik': 'Inverter',
            'omnik_dsmr': 'Omnik_data_logger',
            'dsmr': 'DSMR_electicity_meter',
            'dsmr_gas': 'DSMR_gasmeter'
            }
        return asset_classes

    def _init_attribute_set(self, attribute, asset_class, type='string'):
        if attribute not in self.config.attributes:
            self.config.attributes[attribute] = {}
        if not self.config.has_option('attributes', f'{attribute}.{asset_class}'):
            return
        if type == 'list':
            self.config.attributes[attribute][asset_class] = self.config.getlist(
                'attributes', f'asset.{asset_class}')
        else:
            self.config.attributes[attribute][asset_class] = self.config.get('attributes', f'{attribute}.{asset_class}')

    def _read_attributes(self):
        asset_classes = self._init_attribute_dict()
        try:
            if self.config.has_option('attributes', 'asset_classes'):
                asset_classes += list(set(self.config.getlist('attributes', 'asset_classes')) - set(asset_classes))
            for asset_class in asset_classes:
                self._init_attribute_set('asset', asset_class, 'list')
                self._init_attribute_set('model', asset_class)
                self._init_attribute_set('mf', asset_class)
                self._init_attribute_set('identifier', asset_class)
                self._init_attribute_set('devicename', asset_class)
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR",
                                f"Error reading attributes from config. Error: {e.args}")
            raise e
            return False
        return True

    def terminate(self):
        # Cleanup Client
        self._terminate_client()
        # Cleanup Output plugins
        self._terminate_output_plugins()
        # Terminate DSMR
        if self.dsmr:
            self.dsmr.terminate()

    def _init_client(self):
        # For now the default client is not set by default, client should be configured in the config
        # options are: localproxy, solarmanpv and omnikportal
        self.client_module = self.config.get('plugins', 'client', None)
        if not self.client_module:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "ERROR", "No 'client' parameter configured under the [plugin] section.")
            sys.exit(1)

        hybridlogger.ha_log(self.logger, self.hass_api, "INFO", f"Initializing client : {self.client_module}.")
        # Try to load client
        try:
            Client.logger = self.logger
            Client.config = self.config
            Client.hass_api = self.hass_api
            # __import__(self.client_module)
            spec = importlib.util.find_spec(self.client_module)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            self.client = Client.client[0]
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "ERROR", f"Client '{self.client_module}' could not be initialized. Error: {e}")
            sys.exit(1)

    def _terminate_client(self):
        while Client.client:
            Client.client.pop(0).terminate()

    def _init_output_plugins(self):
        self.plugins = self.config.getlist('plugins', 'output', fallback=[''])
        if self.plugins and self.plugins[0]:
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO", f"Plugins enabled: {self.plugins}.")
        else:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "WARNING", "No output plugins configured! Monitoring only.")
        # Import output plugins
        if self.plugins and self.plugins[0]:
            Plugin.logger = self.logger
            Plugin.config = self.config
            Plugin.hass_api = self.hass_api

            for plugin in self.plugins:
                spec = importlib.util.find_spec(plugin)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                # importlib.import_module(plugin)

    def _terminate_output_plugins(self):
        while Plugin.plugins:
            Plugin.plugins.pop(0).terminate()

    def _init_dsmr(self):
        terminals = self.config.getlist('dsmr', 'terminals', fallback=[])
        if terminals and terminals[0]:
            self.dsmr = DSRM()
            self.dsmr.config = self.config
            self.dsmr.logger = self.logger
            self.dsmr.hass_api = self.hass_api
            self.dsmr.datalogger = self
            self.dsmr.initialize(terminals=terminals, dsmr_callback=self.dsmr_callback)
        else:
            self.dsmr = None
        self.dsmr_data = {}

    def _get_dsmr_data(self, plant, data):
        # if dsmr measurements are not enabled then return
        if not self.dsmr:
            return
        # Try to merge with the latest dsmr data available
        complete = False
        tries = 20
        while tries > 0:
            self.dsmr_access.acquire()
            if plant in self.dsmr_data:
                # Insert raw DSMR data
                data.update(self._dsmr_cache(plant, data['last_update']))
                # Insert calculated netto values (solar - net)
                self._calculate_consumption(data)
                complete = True
            self.dsmr_access.release()
            tries -= 1
            if not complete:
                time.sleep(1)
            else:
                break
        if not tries:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "WARNING", f"Could not combine DSMR data for plant {plant}. "
                                "Did you configure plant_id at your dsmr terminal config?")

    def _calculate_consumption(self, data):
        # Calculate cumulative energy direct used
        data['energy_direct_use'] = data['total_energy_recalc'] - data['total_energy_offset'] - data['energy_delivered_net']
        data['energy_used'] = data['energy_used_net'] + data['energy_direct_use']  # use as v3 * 1000 for pvoutput
        # Calculate the actual power used
        if 'current_power' in data:
            data['power_direct_use'] = data['current_power'] - (data['CURRENT_ELECTRICITY_DELIVERY'] * Decimal('1000'))
            # Should not be negative, but timestamps can be out of sync.
            if data['power_direct_use'] < Decimal('0'):
                data['power_direct_use'] = Decimal('0')
            data['power_consumption'] = data['CURRENT_ELECTRICITY_USAGE'] * Decimal('1000') + data['power_direct_use']
        else:
            data['power_consumption'] = data['CURRENT_ELECTRICITY_USAGE'] * Decimal('1000')
        data['last_update_calc'] = time.time()

    def dsmr_callback(self, terminal, dsmr_message):
        # print(f"{dsmr_message['gas_consumption_totall']} =
        # m3 {dsmr_message['gas_consumption_hour']} m3/h - {dsmr_message['timestamp_gas']}")
        if dsmr_message['plant_id']:
            plant_id = dsmr_message.pop('plant_id')
            # Get last total energy form cache (if available) and ad to the dataset
            # dsmr_message['total_energy_cache'] = self.total_energy(plant_id)
            self.dsmr_access.acquire()
            self._dsmr_cache_update(plant_id, dsmr_message)
            # self.dsmr_data[plant_id] = deepcopy(dsmr_message)
            self.dsmr_access.release()
            if plant_id in self.plant_update:
                self._proces_pushed_net_event(plant_id, dsmr_message)
        else:
            self._proces_pushed_net_event(None, dsmr_message)

    def _dsmr_cache_update(self, plant_id, dsmr_message):
        if plant_id not in self.dsmr_data:
            self.dsmr_data[plant_id] = list()
        if len(self.dsmr_data[plant_id]) == 10:
            self.dsmr_data[plant_id].pop(0)
        self.dsmr_data[plant_id].append(dsmr_message)

    def _dsmr_cache(self, plant_id, timestamp):
        max_delta = -1
        best_data = None
        for data in reversed(self.dsmr_data[plant_id]):
            dsmr_ts = data['timestamp']
            delta = abs(timestamp - dsmr_ts)
            if delta < max_delta or max_delta < 0:
                max_delta = delta
                best_data = data
        return best_data

    def _get_interval(self):
        if self.config.has_option('default', 'interval'):
            return int(self.config.get('default', 'interval'))
        else:
            return 0

    def _get_interval_aggregated(self):
        if self.config.has_option('default', 'interval_aggregated'):
            return int(self.config.get('default', 'interval_aggregated'))
        else:
            return 300

    def _validate_user_login(self):
        self.omnik_api_level = 0
        try:
            self.client.initialize()
            # Logged on
            self.omnik_api_level = 1
        except requests.exceptions.RequestException as err:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "WARNING", f"Request error during account validation omnik portal: {err}")
        except requests.exceptions.HTTPError as errh:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "WARNING", f"HTTP error during account validation omnik portal: {errh}")
        except requests.exceptions.ConnectionError as errc:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "WARNING", f"Connection error during account validation omnik portal: {errc}")
        except requests.exceptions.Timeout as errt:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "WARNING", f"Timeout error during account validation omnik portal: {errt}")
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", e)

    def _logon(self):
        # Log on to the omnik portal to enable fetching plant's
        if (self.omnik_api_level == 0):
            self._validate_user_login()
            if (self.omnik_api_level == 0):
                return False
        return True

    def _fetch_plants(self):
        # Fetch te plant's available for the account
        if (not self.plant_update or self.omnik_api_level == 1):
            try:
                plants = self.client.getPlants()
                for pid in plants:
                    self.plant_update[str(pid['plant_id'])] = self.last_update_time
                self.omnik_api_level = 2
            except requests.exceptions.RequestException as err:
                hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"Request error: {err}")
                self.omnik_api_level = 0
                return False
            except requests.exceptions.HTTPError as errh:
                hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"HTTP error: {errh}")
                self.omnik_api_level = 0
                return False
            except requests.exceptions.ConnectionError as errc:
                hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"Connection error: {errc}")
                self.omnik_api_level = 0
                return False
            except requests.exceptions.Timeout as errt:
                hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"Timeout error: {errt}")
                self.omnik_api_level = 0
                return False
            except Exception as e:
                hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", e)
                self.omnik_api_level = 0
                return False
        return True

    def _listen_for_update(self):
        data = self.client.getPlantData()
        return data

    def _process_received_update(self, data, netdata=False, plant=None):
        # Get the report time
        newreporttime = datetime.fromtimestamp(data['last_update']).astimezone(timezone.utc)
        if not plant:
            plant = data['plant_id']
        if netdata:
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                f"Net data update for plant '{plant}' update at UTC {newreporttime}")
        else:
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                f"Update for plant {plant} update at UTC {newreporttime}")

        return plant

    def _sundown_reset_power(self, data):
        if self.client.use_timer:
            if self.sundown:
                data['current_power'] = Decimal('0')

    def _fetch_update(self, plant):
        try:
            data = self.client.getPlantData(plant)
            if data:
                self._sundown_reset_power(data)
                # Get the actual report time from the omnik portal
                newreporttime = datetime.fromtimestamp(data['last_update']).astimezone(timezone.utc)
                # Only proces updates that occured after we started or start a single measurement (TODO)
                if (newreporttime > self.plant_update[plant] or not self.every or self.sundown):
                    hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                        f"Update for plant {plant} update at UTC {newreporttime}")
                    # the newest plant_update time will be used as a baseline to program the timer
                    self.last_update_time = newreporttime
                    # Store the plant_update time for each indiviual plant,
                    # update only if there is a new report available to avoid duplicates
                    self.plant_update[plant] = newreporttime
                    data['plant_id'] = plant
                    return data
                else:
                    hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                        f'No recent report update to process. Last report at UTC {newreporttime}')
                    return None

        except requests.exceptions.RequestException as err:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"Request error: {err}")
            self.omnik_api_level = 0
            # Abort retry later
            return None
        except requests.exceptions.HTTPError as errh:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"HTTP error: {errh}")
            self.omnik_api_level = 0
            # Abort retry later
            return None
        except requests.exceptions.ConnectionError as errc:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"Connection error: {errc}")
            self.omnik_api_level = 0
            # Abort retry later
            return None
        except requests.exceptions.Timeout as errt:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"Timeout error: {errt}")
            self.omnik_api_level = 0
            # Abort retry later
            return None
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", f"Unexpected error during data handling: {e}")
            self.omnik_api_level = 0
            # Abort retry later
            return None

    def _validate_field(self, data, field, fallback):
        # validate existance of field in dict
        # provide a default value if it is not available
        if data:
            if field not in data:
                data[field] = fallback

    def _validate_client_data(self, plant, data):
        # Insert defaults for missing fields
        # data['data']['plant_id']
        self._validate_field(data, 'inverter', self.config.get(f"plant.{plant}", 'inverter_sn', 'n/a'))
        return data

    def _output_update(self, plant, data):
        # Insert dummy data for fields that have not been supplied by the client
        data = self._validate_client_data(plant, data)
        # Process for each plugin, but only when valid
        for plugin in Plugin.plugins:
            if not plugin.process_aggregates or 'sys_id' in data:
                hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG",
                                    f"Trigger plugin '{getattr(plugin, 'name')}'.")
                plugin.process(msg=data)

    def _output_update_aggregated_data(self, plant, data):
        # Insert dummy data for fields that have not been supplied by the client
        data = self._validate_client_data(plant, data)
        # Process for each plugin, but only when valid
        for plugin in Plugin.plugins:
            if plugin.process_aggregates:
                hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG",
                                    f"Trigger plugin '{getattr(plugin, 'name')}' with aggregated data.")
                plugin.process(msg=data)

    def _init_aggregated_data(self, aggregated_data, data, sys_id):
        if not aggregated_data:
            # Initialize
            aggregated_data['sys_id'] = sys_id
            self._init_aggregated_data_field(aggregated_data, data, 'last_update')
            self._init_aggregated_data_field(aggregated_data, data, 'current_power')
            self._init_aggregated_data_field(aggregated_data, data, 'today_energy')
            self._init_aggregated_data_field(aggregated_data, data, 'total_energy')
            self._init_aggregated_data_field(aggregated_data, data, 'voltage_ac1')
            self._init_aggregated_data_field(aggregated_data, data, 'voltage_ac2')
            self._init_aggregated_data_field(aggregated_data, data, 'voltage_ac3')
            self._init_aggregated_data_field(aggregated_data, data, 'voltage_ac_max')
            self._init_aggregated_data_field(aggregated_data, data, 'inverter_temperature')
            if self.dsmr:
                self._init_aggregated_data_field(aggregated_data, data, 'power_consumption')
                self._init_aggregated_data_field(aggregated_data, data, 'energy_used')
                self._init_aggregated_data_field(aggregated_data, data, 'energy_used_net')
                self._init_aggregated_data_field(aggregated_data, data, 'net_voltage_max')
                self._init_aggregated_data_field(aggregated_data, data, 'INSTANTANEOUS_VOLTAGE_L1')
                self._init_aggregated_data_field(aggregated_data, data, 'INSTANTANEOUS_VOLTAGE_L2')
                self._init_aggregated_data_field(aggregated_data, data, 'INSTANTANEOUS_VOLTAGE_L3')

    def _init_aggregated_data_field(self, aggregated_data, data, field):
        if field in data:
            aggregated_data[field] = Decimal('0')

    def _adapt_max_value(self, aggregated_data, data, field):
        if field in data:
            if data[field] > aggregated_data[field]:
                aggregated_data[field] = data[field]

    def _adapt_add_value(self, aggregated_data, data, field):
        if field in data:
            aggregated_data[field] += data[field]

    def _aggregate_data(self, aggregated_data, data):
        # Check for pvoutput sys_id override in config
        sys_id = int(self.config.get(f"plant.{data['plant_id']}", 'sys_id', '0'))
        global_sys_id = int(self.config.get('output.pvoutput', 'sys_id', '0'))
        if sys_id and (sys_id != global_sys_id):
            # Do no aggegate: add sys_id to data set
            data['sys_id'] = sys_id
        else:
            # Get sys_id from pvoutput section, cannot aggregate without sys_id
            sys_id = global_sys_id
            if sys_id:
                # Aggerate data (initialize dict and set sys_id)
                self._init_aggregated_data(aggregated_data, data, sys_id)
                # Timestamp
                self._adapt_max_value(aggregated_data, data, 'last_update')
                # Add energy
                self._adapt_add_value(aggregated_data, data, 'today_energy')
                self._adapt_add_value(aggregated_data, data, 'total_energy')
                # Add power
                self._adapt_add_value(aggregated_data, data, 'current_power')
                # Max voltage
                self._adapt_max_value(aggregated_data, data, 'voltage_ac_max')
                self._adapt_max_value(aggregated_data, data, 'voltage_ac1')
                self._adapt_max_value(aggregated_data, data, 'voltage_ac2')
                self._adapt_max_value(aggregated_data, data, 'voltage_ac3')
                self._adapt_max_value(aggregated_data, data, 'net_voltage_max')
                self._adapt_max_value(aggregated_data, data, 'INSTANTANEOUS_VOLTAGE_L1')
                self._adapt_max_value(aggregated_data, data, 'INSTANTANEOUS_VOLTAGE_L2')
                self._adapt_max_value(aggregated_data, data, 'INSTANTANEOUS_VOLTAGE_L3')
                # Max inverter temperature
                self._adapt_max_value(aggregated_data, data, 'inverter_temperature')
            if sys_id and self.dsmr:
                self._adapt_add_value(aggregated_data, data, 'power_consumption')
                self._adapt_add_value(aggregated_data, data, 'energy_used_net')
                self._adapt_add_value(aggregated_data, data, 'energy_used')

    def _digitize(self, data):
        digitize_fields = [
            "current_power",
            "today_energy",
            "total_energy",
            "inverter_temperature",
            "current_ac1",
            "current_ac2",
            "current_ac3",
            "voltage_ac1",
            "voltage_ac2",
            "voltage_ac3",
            "voltage_ac_max",
            "frequency_ac1",
            "frequency_ac2",
            "frequency_ac3",
            "power_ac1",
            "power_ac2",
            "power_ac3",
            "current_pv1",
            "current_pv2",
            "current_pv3",
            "voltage_pv1",
            "voltage_pv2",
            "voltage_pv3",
            "power_pv1",
            "power_pv2",
            "power_pv3",
            "current_power_pv",
            "operation_hours"
            ]
        for field in data:
            if field in digitize_fields:
                data[field] = Decimal(f'{data[field]}')
        # Re calculate total enery for today accuracy (TODO per plant_id!) update after every measurement
        # store in cache and make persistant
        # Also store the last current power to the cache
        data['total_energy_recalc'] = self.total_energy(data['plant_id'],
                                                        today_energy=data['today_energy'],
                                                        total_energy=data['total_energy'],
                                                        current_power=data['current_power']
                                                        )

    def _load_persistant_cache(self):
        try:
            with open(self.persistant_cache_file) as total_energy_cache:
                file_cache = json.load(total_energy_cache)
                for item in file_cache:
                    self.cache[item] = Decimal(f"{file_cache[item]}")
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                f"Using energy cache file '{self.persistant_cache_file}'.")
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                f"Cache file '{self.persistant_cache_file}' was not used previously. Error: {e.args}")

    def _update_persistant_cache(self):
        try:
            file_cache = {}
            for item in self.cache:
                file_cache[item] = float(self.cache[item])
            with open(self.persistant_cache_file, 'w') as json_file_config:
                json.dump(file_cache, json_file_config)
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR",
                                f"Cache file '{self.persistant_cache_file}' can not be written! Error: {e.args}")

    def last_current_power(self, plant):
        last_current_power = f'{plant}.last_current_power'
        if last_current_power not in self.cache:
            return Decimal('0')
        else:
            return self.cache[last_current_power]

    def total_energy(self, plant, today_energy=None, total_energy=None, current_power=None, lifetime=True):
        # start_total_energy_key = f'{plant}.start_total_energy'
        last_total_energy = f'{plant}.last_total_energy'
        last_today_energy = f'{plant}.last_today_energy'
        last_current_power = f'{plant}.last_current_power'
        # if total energy is supplied, update cache directly
        if total_energy:
            self.cache[last_total_energy] = total_energy
            self.cache[last_today_energy] = today_energy
            self.cache[last_current_power] = current_power
            self._update_persistant_cache()

        if plant not in self.start_total_energy:
            if total_energy:
                self.start_total_energy[plant] = total_energy - today_energy
                return self.start_total_energy[plant] + today_energy if lifetime else today_energy
            else:
                # No accurate total energy is available, use total_energy cache
                if last_today_energy in self.cache and last_total_energy in self.cache:
                    self.start_total_energy[plant] = self.cache[last_total_energy] - self.cache[last_today_energy]
                    return self.start_total_energy[plant] + self.cache[last_today_energy] if lifetime \
                        else self.cache[last_today_energy]
                else:
                    return None
        else:
            if total_energy and not today_energy:
                self.start_total_energy[plant] = total_energy
                return total_energy
            elif today_energy:
                return self.start_total_energy[plant] + today_energy
            else:
                if last_today_energy in self.cache and last_total_energy in self.cache:
                    return self.start_total_energy[plant] + self.cache[last_today_energy] if lifetime \
                        else self.cache[last_today_energy]
                else:
                    return None

    def _sunshine_check(self):
        self.sundown = not self.dl.sun_shine()
        if self.sundown:
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                f"No sunshine postponing till down next dawn {self.dl.next_dawn}.")
            # Send 0 Watt update
            return self.dl.next_dawn + timedelta(minutes=10)
        else:
            # return the last report time return value, but not when there is no sun
            return self.last_update_time

    def _proces_pushed_net_event(self, plant_id, dsmr_message):

        # ts_midnight = datetime.combine(date.today(), datetime.min.time()).timestamp()
        data = deepcopy(dsmr_message)
        dsmr_timestamp = data['timestamp']
        # if (dsmr_timestamp - last_update) > (self.every + 10) and dsmr_timestamp > self.pasttime:
        if dsmr_timestamp > self.pasttime and plant_id and self.total_energy(plant_id):
            # Process independent net data for aggregated clients with regards of rate limits
            # data['current_power'] = Decimal('0')
            # Omnik related fields need 'last_update' timestamp
            data['last_update'] = dsmr_timestamp
            data['plant_id'] = plant_id
            data['total_energy_recalc'] = self.total_energy(plant_id)
            data['today_energy'] = self.total_energy(plant_id, lifetime=False)
            self._calculate_consumption(data)
            data['total_energy'] = data.pop('total_energy_recalc')
            aggegated_data = {}
            self._aggregate_data(aggegated_data, data)
            # Do not publish the current power
            # set next time block for update
            self.pasttime = dsmr_timestamp - (dsmr_timestamp % self.interval_aggregated) + self.interval_aggregated
            # set net updates out for aggegated clients (pvoutput)
            last_update = datetime.timestamp(self.plant_update[plant_id])
            if (dsmr_timestamp - last_update) > (self.every + 10) or not self.last_current_power(plant_id):
                self._process_received_update(aggegated_data, netdata=True, plant=plant_id)
                self._output_update_aggregated_data(plant_id, aggegated_data)
        # TODO process net update for other clients
        data2 = deepcopy(dsmr_message)
        data2['plant_id'] = plant_id
        self._validate_client_data(plant_id, data2)
        # data['last_update'] = dsmr_timestamp
        # data['net_update'] = True
        # dsmr_timestamp = data.pop('timestamp')
        # data['last_update'] = dsmr_timestamp
        # plant = self._process_received_update(data)
        # Set last update time for plant
        # self.plant_update[plant] = datetime.fromtimestamp(data['last_update'])
        # Digitize data
        # self._digitize(data)
        # Get dsmr data
        # self._get_dsmr_data(plant, data)
        # Assemble aggegated data
        # export the data to the output plugins
        self._output_update(plant_id, data2)
        # Process aggregated

    def _process_timed_event(self):
        # For portal clients and tcpclient with a fixed interval
        # Returns the date time of the last update (time behind) or a time in te future when postposing updates
        # Do a Sunshine check when a Timer is used
        next_report_at = self._sunshine_check()

        # Check for login, if needed. Return None on failure
        if not self._logon():
            # Logon failed
            # Sun is down, skip this measurement and wait till dawn
            return next_report_at if self.sundown else None

        # Caching of plant id's, if needed. Return None on failure
        if not self._fetch_plants():
            # Sun is down, skip this measurement and wait till dawn
            return next_report_at if self.sundown else None

        # Process data reports for each plant
        if self.omnik_api_level == 2:
            aggegated_data = {}
            for plant in self.plant_update:
                # Try getting a new update or wait for logging event
                data = self._fetch_update(plant)
                if data:
                    # A new last update time was set
                    if self.plant_update[plant] > next_report_at:
                        next_report_at = self.plant_update[plant]
                    # Digitize data
                    self._digitize(data)
                    # Get dsmr data
                    self._get_dsmr_data(plant, data)
                    # Assemble aggegated data
                    self._aggregate_data(aggegated_data, data)
                    # export the data to the output plugins
                    self._output_update(plant, data)
            # Process aggregated data
            if aggegated_data:
                self._output_update_aggregated_data(plant, aggegated_data)
                hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", 'Aggregated data processed.')

        # Finish datalogging process
        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", 'Timed data logging processed')

        # Return the the time of the latest report received
        return next_report_at

    def _process_pushed_event(self):
        # For localproxy client
        # Returns the date time of the last update
        next_report_at = None

        # Check for login, if needed. Return None on failure
        if not self._logon():
            # Skip this measurement, retry later
            return None

        # Caching of plant id's, if needed. Return None on failure
        if not self._fetch_plants():
            # Skip this measurement, retry later
            return None

        # Process data reports for each plant
        if self.omnik_api_level == 2:
            # Wait for new data logging event
            data = self._listen_for_update()
            if data:
                aggegated_data = {}
                plant = self._process_received_update(data)
                # Set last update time for plant
                self.plant_update[plant] = datetime.fromtimestamp(data['last_update'])
                # Digitize data
                self._digitize(data)
                # Get dsmr data
                self._get_dsmr_data(plant, data)
                # Assemble aggegated data
                self._aggregate_data(aggegated_data, data)
                # export the data to the output plugins
                self._output_update(plant, data)
                # Process aggregated
                self._output_update_aggregated_data(plant, aggegated_data)

        # Finish datalogging process
        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", 'Pushed logging processed')

        # Return the the time of the latest report received
        return next_report_at

    def process(self):
        # Returns the date time of the last update (time behind) or a time in te future when postposing updates
        # Do a Sunshine check when a Timer is used
        if self.client.use_timer:
            return self._process_timed_event()
        else:
            # Return None when an error occurs
            return self._process_pushed_event()

    @staticmethod
    def __expand_path(path):
        """
        Expand relative path to absolute path.
        Args:
            path: file path
        Returns: absolute path to file
        """
        if os.path.isabs(path):
            return path
        else:
            return os.path.dirname(os.path.abspath(__file__)) + "/" + path



class BaseClient(type):
    def __init__(cls, name, bases, attrs):
        super(BaseClient, cls).__init__(name)
        if not hasattr(cls, 'client'):
            cls.client = []
        else:
            cls.register(cls)  # Called when a client class is imported

    def register(cls, client):
        cls.client.append(client())


class Client(object, metaclass=BaseClient):

    config = None
    logger = None
    hass_api = None
    use_timer = True

    def initialize(self):
        pass

    def terminate(self):
        pass



from omnik.ha_logger import hybridlogger
import omnik.InverterMsg
from omnik.plugin_client import Client
from omnik.plugin_localproxy import LocalProxyPlugin
import time
import threading
import importlib


class LocalProxy(Client):

    # TODO implementation of logging using a local proxy using https://github.com/Woutrrr/Omnik-Data-Logger
    # and https://github.com/t3kpunk/Omniksol-PV-Logger

    def __init__(self):
        super().__init__()

        hybridlogger.ha_log(self.logger, self.hass_api, "INFO", "Client enabled: localproxy")

        # Switch of the repeated timer
        self.use_timer = False
        # Create a semaphore for unique access to the processing loop
        self.semaphore = threading.Semaphore()
        self.msgevent = threading.Event()
        self.msg = {
            'isSet': False,
            'data': None,
            'plugin': None
            }

        # Get plant_id_list
        self.plant_id_list = self.config.getlist('client.localproxy', 'plant_id_list', fallback=[''])
        if not self.plant_id_list or self.plant_id_list[0] == '':
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", "plant_id_list was not specified for client localproxy")
            raise Exception('plant_id_list was not specified')
        # Import localproxy plugins
        self.plugins = self.config.getlist('plugins', 'localproxy', fallback=[''])
        if not self.plugins[0] == '':
            LocalProxyPlugin.semaphore = self.semaphore
            LocalProxyPlugin.logger = self.logger
            LocalProxyPlugin.config = self.config
            LocalProxyPlugin.hass_api = self.hass_api
            LocalProxyPlugin.client = self
            for plugin in self.plugins:
                # __import__(plugin)
                spec = importlib.util.find_spec(plugin)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
        else:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "ERROR", "At least one plugin needs to be configured. "
                                "No 'localproxy' plugins found at the [plugins] section.")
            self.semaphore.release()
            raise Exception('No localproxy plugins were specified.')
        # Initialize dict with inverter info
        self.inverters = {}

    def terminate(self):
        # Stop all plugins
        # Claim the semaphore
        self.semaphore.acquire()
        while LocalProxyPlugin.localproxy_plugins:
            LocalProxyPlugin.localproxy_plugins.pop(0).terminate()
        # Release the semaphore
        self.semaphore.release()

    def getPlants(self):
        # Get the plant neeeded data from config
        data = []
        inverterdata = {}
        for plant in self.plant_id_list:
            inverter_sn = self.config.get(f"plant.{plant}", 'inverter_sn', fallback=None)
            if not inverter_sn:
                hybridlogger.ha_log(self.logger, self.hass_api,
                                    "ERROR", "inverter_sn (The serial number of the inverter) for "
                                    f"plant {plant} was not specified for [TCPclient]")
                raise Exception('inverter_sn (a serial number of the inverter) was not specified')
            inverterdata = {
                "inverter_sn": inverter_sn
                }
            self.inverters[plant] = inverterdata
            data.append({'plant_id': plant})

        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", f"plant list from config {data}")

        # The config was read, start listening
        # Claim the semaphore
        self.semaphore.acquire()
        for plugin in LocalProxyPlugin.localproxy_plugins:
            plugin.listen()
        # Release the semaphore
        self.semaphore.release()
        return data

    def getPlantData(self, plant_id=None):
        # Wait here for a message
        self.msgevent.wait()
        # Claim the semaphore
        valid = False
        data = None
        self.semaphore.acquire()
        try:
            if self.msg['isSet']:
                data = {}
                if self.msg['data']:
                    if len(self.msg['data']) == 128:
                        inverterMsg = omnik.InverterMsg.InverterMsg(self.msg['data'])
                        serialnr = inverterMsg.getID()
                        # lookup plant_id to see it is in the plants list
                        for plant in self.plant_id_list:
                            if self.inverters[plant]['inverter_sn'] == serialnr:
                                data['plant_id'] = plant
                                valid = True
                if valid:
                    # Get the data from the received message
                    inverterMsg.FetchDataDict(data)
                    hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                        f"New message received from inverter '{serialnr}. Plugin: {self.msg['plugin']}'")
                else:
                    hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", "Invalid response, ignoring message")
                    data = None

                # Reset isSet flag
                self.msg['isSet'] = False

        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"Error occured while processing message. Error: {e}")
        # Release the semaphore and event
        self.semaphore.release()
        self.msgevent.clear()
        # take some sleep to relax
        time.sleep(1)
        return data



#! /usr/bin/env python3
#
# This script can be used to intercept and forward the inverter log messages.
# The script has been tested with python versions 2.7 and 3.8
# The data can be forwarded to MQTT to process using omnikdatalogger with the halogger client.
# To intercept on your local network:
# Route 176.58.117.69/32 to the device that runs this logger and listen at port 10004
# On the device NAT the poblic address to the private address (e.g. 192.168.x.x ) using iptables:
# iptables -t nat -A PREROUTING -p tcp -d 176.58.117.69 --dport 10004 -j DNAT --to-destination 192.168.x.x:10004
# iptables -t nat -A OUTPUT -p tcp -d 176.58.117.69 -j DNAT --to-destination 192.168.x.x
#

import socket
import socketserver
import argparse
import configparser
import yaml
import threading
import signal
import json
import binascii
import os
import paho.mqtt.client as mqttclient
import uuid
import logging
import datetime
import time

__version__ = '1.2.6'
listenaddress = b'127.0.0.1'                       # Default listenaddress
listenport = 10004                                 # Make sure your firewall enables you listening at this port
# There is no need to change this if this proxy must log your data directly to the Omnik/SolarmanPV servers
omnikloggerpublicaddress = b'176.58.117.69'        # If you have an aditional omniklogger, change the ip to enlarge the chain
omnikloggerdestport = 10004                        # This is the port the omnik/SolarmanPV datalogger server listens to
STATUS_ON = 'ON'
STATUS_OFF = 'OFF'
CHECK_STATUS_INTERVAL = 60
INVERTER_MAX_IDLE_TIME = 6

global stopflag
logger = logging.getLogger(__name__)


# Generic signal handler
def signal_handler(signal, frame):
    global stopflag
    logger.debug("Signal {:0d} received. Setting stopflag.".format(signal))
    stopflag = True


class ProxyServer(threading.Thread):

    def __init__(self, args=[], kwargs={}):
        threading.Thread.__init__(self)
        if args.mqtt_host:
            logger.info('Enabling MQTT forward to {0}'.format(args.mqtt_host))
            RequestHandler.mqttfw = mqtt(args)
            # Check every 60 sec if there is a valid update status aged less than 30 minutes
            self.statustimer = threading.Timer(CHECK_STATUS_INTERVAL, self.check_status)
            self.statustimer.start()

        # Create tcp server
        self.tcpServer = socketserver.TCPServer((args.listenaddress, args.listenport), RequestHandler)

    def check_status(self):
        for serial in RequestHandler.lastupdate:
            if RequestHandler.lastupdate[serial] + datetime.timedelta(minutes=INVERTER_MAX_IDLE_TIME) < \
                    datetime.datetime.now():
                RequestHandler.status[serial] = STATUS_OFF
                RequestHandler.mqttfw.mqttforward('',
                                                  serial, RequestHandler.status[serial])
        # Restart timer
        self.statustimer = threading.Timer(CHECK_STATUS_INTERVAL, self.check_status)
        self.statustimer.start()

    def cancel(self):
        # Stop the loop
        self.tcpServer.shutdown()
        if RequestHandler.mqttfw:
            self.statustimer.cancel()
            RequestHandler.mqttfw.close()

    def run(self):
        # Try to run TCP server
        logger.info('Start in Omnik proxy server. Listening to {0}:{1}'.format(args.listenaddress, args.listenport))
        self.tcpServer.serve_forever()


class RequestHandler(socketserver.BaseRequestHandler):

    mqttfw = None
    status = {}
    lastupdate = {}

    def handle(self):
        msg = self.request.recv(1024)

        if len(msg) >= 99:
            self._processmsg(msg)

    def _processmsg(self, msg):
        rawserial = str(msg[15:31].decode())
        # obj dLog transforms de raw data from the photovoltaic Systems converter
        valid = False
        if rawserial in args.serialnumber:
            # TODO: Dit gaat niet werken zo
            RequestHandler.status[rawserial] = STATUS_ON
            RequestHandler.lastupdate[rawserial] = datetime.datetime.now()
            valid = True
        if valid:
            # Process data
            logger.info("Processing message for inverter '{0}'".format(rawserial))
            self.forwardstate(msg, rawserial, self.status[rawserial])
        else:
            logger.warning("Ignoring received data!")

    def forwardstate(self, msg, serial, status):
        # Forward data to datalogger and MQTT
        # TODO
        # Forward logger message to MTTQ if host is defined
        if RequestHandler.mqttfw:
            RequestHandler.mqttfw.mqttforward(json.dumps(str(binascii.b2a_base64(msg), encoding='ascii')),
                                              serial, status)
        # Forward data to proxy or Omnik datalogger
        if args.omniklogger:
            self.fwthread = tcpforward()
            self.fwthread.forward(msg)
            self.fwthread.join(60)
            if self.fwthread.is_alive():
                # Kill the thread if it is still blocking
                logger.warning("Waiting for forward connection takes a long time...")
                self.fwthread.join()


class tcpforward(threading.Thread):

    def forward(self, data):
        threading.Thread.__init__(self, target=self._run)
        self.data = data
        self.start()

    def _run(self):
        loggeraddress = (args.omniklogger, int(args.omnikloggerport))
        self.forwardsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.forwardsock.settimeout(90)
        try:
            self.forwardsock.connect(loggeraddress)
            self.forwardsock.sendall(self.data)
            logger.info('{0} Forwarded to "{1}"'.format(datetime.datetime.now(), args.omniklogger))
            self.forwardsock.close()
        except Exception as e:
            logger.warning("Error forwarding: {0}".format(e))
        finally:
            self.forwardsock.close()


class mqtt(object):

    def __init__(self, args=[], kwargs={}):
        self.mqtt_client_name = args.mqtt_client_name_prefix + uuid.uuid4().hex
        self.mqtt_host = args.mqtt_host
        self.mqtt_port = int(args.mqtt_port)
        self.mqtt_retain = args.mqtt_retain
        if not args.mqtt_username or not args.mqtt_password:
            logger.error("Please specify MQTT username and password in the configuration")
            self.mqtt_client = None
            return
        else:
            self.mqtt_username = args.mqtt_username
            self.mqtt_password = args.mqtt_password

        # mqtt setup
        self.mqtt_client = mqttclient.Client(self.mqtt_client_name)
        self.mqtt_client.on_connect = self._mqtt_on_connect  # bind call back function
        self.mqtt_client.on_disconnect = self._mqtt_on_disconnect  # bind call back function
        # self.mqtt_client.on_message=mqtt_on_message (not used)
        self.mqtt_client.username_pw_set(self.mqtt_username, self.mqtt_password)
        self.mqtt_client.connect(self.mqtt_host, self.mqtt_port)
        # start processing messages
        self.mqtt_client.loop_start()
        # default prefixes
        self.discovery_prefix = args.mqtt_discovery_prefix
        self.device_name = args.mqtt_device_name
        self.logger_sensor_name = args.mqtt_logger_sensor_name

        # lock
        self.lock = threading.Condition(threading.Lock())

    def close(self):
        if self.mqtt_client:
            try:
                self.mqtt_client.disconnect()
            except Exception:
                pass

    def _topics(self):
        topics = {}
        topics['main'] = "{0}/binary_sensor/{1}_{2}".format(self.discovery_prefix, self.logger_sensor_name, self.serial)
        topics['state'] = "{0}/state".format(topics['main'])
        topics['attr'] = "{0}/attr".format(topics['main'])
        topics['config'] = {}
        topics['config']['logger_sensor'] = "{0}/logger_sensor/config".format(topics['main'])
        return topics

    def _device_payload(self):
        # Determine model
        model = "Omnik datalogger proxy"

        # Device payload
        device_pl = {
            "identifiers": ["{0}_{1}".format(self.device_name, self.serial)],
            "name": "{0}_{1}".format(self.device_name, self.serial),
            "mdl": model,
            "mf": 'JBsoft',
            "sw": __version__
            }
        return device_pl

    def _config_payload(self, topics):
        # Get device payload
        device_pl = self._device_payload()
        # Fill config_pl dict
        config_pl = {}
        # current_power
        config_pl['logger_sensor'] = {
            "~": "{0}".format(topics['main']),
            "uniq_id": "{0}_{1}_logger".format(self.device_name, self.serial),
            "name": "{0}".format(self.logger_sensor_name),
            "stat_t": "~/state",
            "json_attr_t": "~/attr",
            "dev_cla": "power",
            "val_tpl": "{{ value_json.state }}",
            "dev": device_pl
            }
        return config_pl

    def _value_payload(self):
        value_pl = {"state": self.status}
        return value_pl

    def _attribute_payload(self):
        attr_pl = {
            "inverter": self.serial,
            "data": self.data,
            "timestamp": time.time(),
            "last_update": "{0} {1}".format(self.reporttime.strftime('%Y-%m-%d'), self.reporttime.strftime('%H:%M:%S'))
            }
        return attr_pl

    def _publish_config(self, topics, config_pl, entity):
        try:
            # publish config
            if self.mqtt_client.publish(topics['config'][entity], json.dumps(config_pl[entity]), retain=self.mqtt_retain):
                logger.debug("Publishing config {0} successful.".format(entity))
                return True
            else:
                logger.warning("Publishing config {0} failed!".format(entity))
                return False
        except Exception as e:
            logger.error("Unhandled error publishing config for entity {0}: {1}".format(entity, e))
            return False

    def _publish_attributes(self, topics, attr_pl):
        try:
            # publish attributes
            if self.mqtt_client.publish(topics['attr'], json.dumps(attr_pl), retain=self.mqtt_retain):
                logger.debug("Publishing attributes successful.")
                return True
            else:
                logger.warning("Publishing attributes failed!")
                return False
        except Exception as e:
            logger.error("Unhandled error publishing attributes: {0}".format(e))
            return False

    def _publish_state(self, topics, value_pl):
        try:
            # publish state
            if self.mqtt_client.publish(topics['state'], json.dumps(value_pl), retain=self.mqtt_retain):
                logger.debug("Publishing state {0} successful.".format(json.dumps(value_pl)))
                return True
            else:
                logger.warning("Publishing state {0} failed!".format(json.dumps(value_pl)))
                return False
        except Exception as e:
            logger.error("Unhandled error publishing states: {0}".format(e))
            return False

    def mqttforward(self, data, serial, status):
        # Decode data: base64.b64decode(json.loads(d, encoding='UTF-8'))
        self.lock.acquire()
        self.data = data
        self.serial = serial
        self.status = status
        self.reporttime = datetime.datetime.now()
        # Get topics
        topics = self._topics()
        # publish attributes
        attr_pl = self._attribute_payload()
        self._publish_attributes(topics, attr_pl)

        # Publish config
        config_pl = self._config_payload(topics)
        for entity in config_pl:
            self._publish_config(topics, config_pl, entity)

        # publish state
        value_pl = self._value_payload()
        if self._publish_state(topics, value_pl):
            logger.info('{0} Message forwarded to MQTT service at "{1}"'.format(datetime.datetime.now(), args.mqtt_host))

        self.lock.release()

    def _mqtt_on_connect(self, client, userdata, flags, rc=0):
        if rc == 0:
            logger.info("MQTT connected")
            # subscribe listening (not used)

    def _mqtt_on_disconnect(self, client, userdata, flags, rc=0):
        if rc == 0:
            logger.info("MQTT disconnected")


def main(args):
    global stopflag
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    proxy = ProxyServer(args)
    proxy.start()
    try:
        while not stopflag and proxy.is_alive():
            proxy.join(1)
    except KeyboardInterrupt:
        pass
    proxy.cancel()
    logger.info("Stopping threads...")
    proxy.join()


def get_yaml_settings(args):
    with open(args.settings, 'r') as stream:
        try:
            settings = yaml.safe_load(stream)
            if args.section in settings:
                index = args.section
            else:
                for key in settings:
                    index = key
                    break
            logger.info("Using section '{1}' from config file '{0}'".
                        format(args.settings, index))
            return settings[index]
        except yaml.YAMLError as exc:
            logger.error("YAML config file '{0}' could not be parsed. Error: {1}".
                         format(args.settings, exc))
            os.sys.exit(1)


def get_yaml_setting(settings, section, key, default):
    if not settings:
        return default
    elif section not in settings:
        return default
    elif key not in settings[section]:
        return default
    else:
        return settings[section][key]


if __name__ == '__main__':
    stopflag = False
    home = os.path.expanduser('~')

    parser = argparse.ArgumentParser()
    parser.add_argument('--settings', default=os.path.join(home, '.omnik/config.yaml'),
                        help='Path to .yaml configuration file', metavar="FILE")
    parser.add_argument('--section', default=None,
                        help='Section to .yaml configuration file to use. Defaults to the first section found.')
    parser.add_argument('--config', default=os.path.join(home, '.omnik/config.ini'),
                        help='Path to configuration file (ini) (DECREPATED!)', metavar="FILE")
    parser.add_argument('--serialnumber', default=None, nargs='+',
                        help='The serial number(s) of your inverter (required)')
    parser.add_argument('--loglevel', default='INFO',
                        help='The basic loglevel [DEBUG, INFO, WARNING, ERROR, CRITICAL]')
    parser.add_argument('--listenaddress', default=listenaddress,
                        help='A local available address to listen to')
    parser.add_argument('--listenport', default=listenport, type=int,
                        help='The local port to listen to')
    parser.add_argument('--omniklogger', default=omnikloggerpublicaddress,
                        help='Forward to an address omnik/SolarmanPV datalogger server listens to. '
                             'Set this to {0} as final forwarder.'.format(omnikloggerpublicaddress))
    parser.add_argument('--omnikloggerport', default=omnikloggerdestport, type=int,
                        help='The port the omnik/SolarmanPV datalogger server listens to')
    parser.add_argument('--mqtt_host', default='',
                        help='The mqtt host to forward processed data to. Config overrides.')
    parser.add_argument('--mqtt_port', default=1883, type=int,
                        help='The mqtt server port. Config overrides.')
    parser.add_argument('--mqtt_retain', default=True, type=bool,
                        help='The mqtt data message retains. Config overrides.')
    parser.add_argument('--mqtt_discovery_prefix', default='homeassistant',
                        help='The mqtt topic prefix (used for MQTT auto discovery). Config overrides.')
    parser.add_argument('--mqtt_client_name_prefix', default='ha-mqtt-omnikloggerproxy',
                        help='The port the omnik/SolarmanPV datalogger server listens to. Config overrides.')
    parser.add_argument('--mqtt_username', default=None,
                        help='The mqtt username. Config overrides.')
    parser.add_argument('--mqtt_password', default=None,
                        help='The mqtt password. Config overrides.')
    parser.add_argument('--mqtt_device_name', default='Datalogger proxy',
                        help='The device name that apears using mqtt autodiscovery (HA). Config overrides.')
    parser.add_argument('--mqtt_logger_sensor_name', default='Datalogger',
                        help='The name of data sensor using mqtt autodiscovery (HA). Config overrides.')
    args = parser.parse_args()
    FORMAT = '%(module)s: %(message)s'
    loglevel = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL
        }

    if os.path.isfile(args.settings):
        settings = get_yaml_settings(args)
        args.mqtt_host = get_yaml_setting(settings, 'output.mqtt', 'host', args.mqtt_host)
        args.mqtt_port = get_yaml_setting(settings, 'output.mqtt', 'port', args.mqtt_port)
        args.mqtt_retain = get_yaml_setting(settings, 'output.mqtt', 'retain', args.mqtt_retain)
        args.mqtt_discovery_prefix = get_yaml_setting(settings, 'output.mqtt',
                                                      'discovery_prefix', args.mqtt_discovery_prefix)
        args.mqtt_client_name_prefix = get_yaml_setting(settings, 'output.mqtt',
                                                        'client_name_prefix', args.mqtt_client_name_prefix)
        args.mqtt_username = get_yaml_setting(settings, 'output.mqtt', 'username', args.mqtt_username)
        args.mqtt_password = get_yaml_setting(settings, 'output.mqtt', 'password', args.mqtt_password)
        args.mqtt_device_name = get_yaml_setting(settings, 'output.mqtt', 'device_name', args.mqtt_device_name)
        args.mqtt_logger_sensor_name = get_yaml_setting(settings, 'output.mqtt',
                                                        'logger_sensor_name', args.mqtt_logger_sensor_name)
        args.serialnumber = get_yaml_setting(settings, 'proxy', 'serialnumber', args.serialnumber)
        args.loglevel = get_yaml_setting(settings, 'proxy', 'loglevel', args.loglevel)
        args.listenaddress = get_yaml_setting(settings, 'proxy', 'listenaddress', args.listenaddress)
        args.listenport = get_yaml_setting(settings, 'proxy', 'listenport', args.listenport)
        args.omniklogger = get_yaml_setting(settings, 'proxy', 'omniklogger', args.omniklogger)
        args.omnikloggerport = get_yaml_setting(settings, 'proxy', 'omnikloggerport', args.omnikloggerport)
    elif os.path.isfile(args.config):
        c = configparser.ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]})
        c.read([args.config], encoding='utf-8')
        args.mqtt_host = c.get('output.mqtt', 'host', fallback=args.mqtt_host)
        args.mqtt_port = c.getint('output.mqtt', 'port', fallback=args.mqtt_port)
        args.mqtt_retain = True if c.get('output.mqtt', 'retain', fallback='true') == 'true' else args.mqtt_retain
        args.mqtt_discovery_prefix = c.get('output.mqtt', 'discovery_prefix ', fallback=args.mqtt_discovery_prefix)
        args.mqtt_client_name_prefix = c.get('output.mqtt', 'client_name_prefix', fallback=args.mqtt_client_name_prefix)
        args.mqtt_username = c.get('output.mqtt', 'username', fallback=args.mqtt_username)
        args.mqtt_password = c.get('output.mqtt', 'password', fallback=args.mqtt_password)
        args.mqtt_device_name = c.get('output.mqtt', 'device_name', fallback=args.mqtt_device_name)
        args.logger_sensor_name = c.get('output.mqtt', 'logger_sensor_name', fallback=args.mqtt_logger_sensor_name)

        # TODO: Try to get from config from no arguments are given. Handle required parameters and defaults
        args.serialnumber = c.getlist('proxy', 'serialnumber', fallback=args.serialnumber)
        args.loglevel = c.get('proxy', 'loglevel', fallback=args.loglevel)
        args.listenaddress = c.get('proxy', 'listenaddress', fallback=args.listenaddress)
        args.listenport = c.getint('proxy', 'listenport', fallback=args.listenport)
        args.omniklogger = c.get('proxy', 'omniklogger', fallback=args.omniklogger)
        args.omnikloggerport = c.getint('proxy', 'omnikloggerport', fallback=args.omnikloggerport)
    if not args.serialnumber:
        parser.print_help()
        os.sys.exit(1)
    logging.basicConfig(level=loglevel[args.loglevel], format=FORMAT)
    main(args)



from omnik.ha_logger import hybridlogger
from omnik.plugin_localproxy import LocalProxyPlugin
import threading
import socketserver
import os


class RequestHandler(socketserver.BaseRequestHandler):

    client = None

    def handle(self):
        data = self.request.recv(1024)

        if self.client and len(data) >= 99:
            # Fill data structure
            self.client.semaphore.acquire()
            self.client.msg['data'] = data
            self.client.msg['isSet'] = True
            self.client.msg['plugin'] = __name__
            self.client.semaphore.release()
            # Trigger processing the message
            self.client.msgevent.set()


class TCPproxy(LocalProxyPlugin):

    '''
    This plugin enables you to listen directly to the logger requests send by the inverter
    Listing can be done by forwarding and NAT to the [tcp_proxy].listen_address:listen_port
    The scripts/proxy/omnikloggerproxy script can also be used to capture and forward to this listener

    This class makes use the following localproxy client objects
    self.client.msg
    self.client.msgevent
    self.client.semaphore
    self.client.inverters
    self.client.plant_id_list
    '''

    def __init__(self):
        super().__init__()
        hybridlogger.ha_log(self.logger, self.hass_api, "INFO", "localproxy client plugin: TCPproxy")
        self.listen_address = self.config.get('client.localproxy.tcp_proxy', 'listen_address', fallback='0.0.0.0')
        self.listen_port = int(self.config.get('client.localproxy.tcp_proxy', 'listen_port', fallback='10004'))
        self.listenaddress = (self.listen_address, int(self.listen_port))
        # Create tcp server
        RequestHandler.client = self.client
        self.tcpServer = socketserver.TCPServer(self.listenaddress, RequestHandler)

    def terminate(self):
        try:
            # Shutting down tcp server
            self.tcpServer.shutdown()
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING",
                                f"Error shutting down tcp_proxy server. Error: {e}.")
            # exit the hard way!
            os.sys.exit(1)

    def listen(self):
        # Start listening thread
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

    def _run(self):
        # TCP listen loop
        try:
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                f"Starting tcp_proxy server. listening at {self.listenaddress[0]}:{self.listenaddress[1]}.")
            self.tcpServer.serve_forever()
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR",
                                f"Error binding to {self.listenaddress}. Error: {e}.")



from omnik.ha_logger import hybridlogger
import omnik.InverterMsg
from omnik.plugin_client import Client
import socket
import binascii


class TCPclient(Client):

    # IMPORTANT : ONLY WIFI Modules with s/n 602xxxxxx and 604xxxxxx support direct access thru port 8899!
    # Based on: https://github.com/woutrrr/omnik-data-logger by Wouter van der Zwan
    # The script was tested only using simulations, let me know if this works for you

    def __init__(self):
        super().__init__()
        hybridlogger.ha_log(self.logger, self.hass_api, "INFO", "Client enabled: TCPclient")

        # Get plant_id_list
        self.plant_id_list = self.config.getlist('client.tcpclient', 'plant_id_list', fallback=[])
        if not self.plant_id_list:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", "plant_id_list was not specified for client TCPclient")
            raise Exception('plant_id_list was not specified')
        # Initialize dict with inverter info
        self.inverters = {}

    def getPlants(self):
        data = []
        inverterdata = {}
        for plant in self.plant_id_list:
            inverter_address = self.config.get(f"plant.{plant}", 'inverter_address', fallback=None)
            if not inverter_address:
                hybridlogger.ha_log(self.logger, self.hass_api, "ERROR",
                                    f"inverter_address for plant {plant} was not specified [TCPclient]")
                raise Exception('an inverter_address was not specified')
            inverter_port = int(self.config.get(f"plant.{plant}", 'inverter_port', fallback='8899'))
            inverter_connection = (inverter_address, int(inverter_port))
            logger_sn = self.config.get(f"plant.{plant}", 'logger_sn', fallback=None)
            if not logger_sn:
                hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", "logger_sn (The serial number of the "
                                    "Wi-Fi datalogger) for plant {plant} was not specified for [TCPclient]")
                raise Exception('logger_sn (a serial number f the Wi-Fi datalogger) was not specified')
            inverter_sn = self.config.get(f"plant.{plant}", 'inverter_sn', fallback=None)
            if not inverter_sn:
                hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", "inverter_sn (The serial number of the inverter) "
                                    "for plant {plant} was not specified for [TCPclient]")
                raise Exception('inverter_sn (a serial number of the inverter) was not specified')
            inverterdata = {
                "inverter_address": inverter_address,
                "inverter_port": inverter_port,
                "logger_sn": int(logger_sn),
                "inverter_sn": inverter_sn,
                "inverter_connection": inverter_connection
                }
            self.inverters[plant] = inverterdata
            data.append({'plant_id': plant})

        hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", f"plant list from config {data}")

        return data

    def getPlantData(self, plant_id):

        valid = False
        data = {}
        # Create request message
        requestmsg = omnik.InverterMsg.request_string(self.inverters[plant_id]['logger_sn'])
        # Create a TCP/IP socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(40)
        try:
            hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", 'Connecting to %s port %s'
                                % self.inverters[plant_id]['inverter_connection'])
            sock.connect(self.inverters[plant_id]['inverter_connection'])
            hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", "Sending request message with logger_sn: "
                                f"{omnik.InverterMsg.request_string(self.inverters[plant_id]['logger_sn'])}")
            # Sending request message
            sock.sendall(requestmsg)
            # Wait for response (listen and block thread)
            rawmsg = sock.recv(129)
            # Validate the data response. Check length and serial number
            if len(rawmsg) >= 99:
                inverterMsg = omnik.InverterMsg.InverterMsg(rawmsg)
                serialnr = inverterMsg.getID()
                # lookup plant_id to see it is in the plants list
                if self.inverters[plant_id]['inverter_sn'] == serialnr:
                    data['plant_id'] = plant_id
                    valid = True
                hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG", f"data size: {len(rawmsg)}.\n"
                                    f"Datadump: {str(binascii.b2a_base64(rawmsg))}")
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"Error getting data from inverter. {e}")
            return None

        if valid:
            # Get the data from the received message
            inverterMsg.FetchDataDict(data)
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO",
                                f"New message received from inverter '{serialnr}'")
        else:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", "Invalid response, ignoring message")
            data = None
        return data



#! /usr/bin/env python3

import os
import signal
import argparse
import configparser
import pathlib
import time
import logging
import yaml
from omnik.ha_logger import hybridlogger
from omnik import RepeatedJob
from omnik.datalogger import DataLogger
try:
    # AppDeamon is only needed when run used with AppDaemon
    import appdaemon.plugins.hass.hassapi as hass
except ImportError:
    # AppDeamon is not needed when run in command mode uses dummy class
    class hass(object):
        Hass = object

logger = logging.getLogger(__name__)

# customize config parser with dict based lookup for AppDaemon and command line options
# has_option and get function have been adapted
ha_args = {}

global stopflag


# Generic signal handler
def signal_handler(signal, frame):
    global stopflag
    logging.debug("Signal {:0d} received. Setting stopflag.".format(signal))
    stopflag = True


class ha_ConfigParser(configparser.ConfigParser):

    def __init__(self, ha_args: dict = None, *args, **kwargs):
        if ha_args is None:
            self.ha_args = {}
        else:
            self.ha_args = ha_args
        super().__init__(*args, **kwargs)

    def has_option(self, section: str, option: str):
        if str(section).lower() == 'default':
            if option in self.ha_args:
                return True
        else:
            if section in self.ha_args:
                if option in self.ha_args[section]:
                    return True
        return super().has_option(section, option)

    def get(self, section, option, fallback=None, **kwargs):
        if str(section).lower() == 'default':
            if option in self.ha_args:
                return self.ha_args.get(option, fallback)
        else:
            if str(section) in self.ha_args:
                if option in self.ha_args[section]:
                    return self.ha_args[section].get(option, fallback)
        try:
            if fallback:
                retval = super().get(section, option, fallback=fallback, **kwargs)
            else:
                retval = super().get(section, option, fallback='', **kwargs)
        except Exception:
            retval = fallback
        return retval

    def getboolean(self, section, option, fallback=None, **kwargs):
        truelist = ['true', '1', 't', 'y', 'yes', 'j', 'ja']
        valstr = str(self.get(section, option, fallback, **kwargs)).lower()
        return (valstr in truelist)

    def getlist(self, section, option, fallback=[], **kwargs):
        if str(section).lower() == 'default':
            if option in self.ha_args:
                return self.ha_args.get(option, fallback)
        else:
            if str(section) in self.ha_args:
                if option in self.ha_args[section]:
                    return self.ha_args[section].get(option, fallback)
        try:
            retval = super().getlist(section, option, fallback=fallback, **kwargs)
        except Exception:
            retval = fallback
            pass
        return retval


# Initialization class for AppDaemon (homeassistant)
class HA_OmnikDataLogger(hass.Hass):

    def initialize(self, *args, **kwargs):
        hybridlogger.ha_log(logger, self, "INFO", "Starting Omnik datalogger...")
        hybridlogger.ha_log(logger, self, "DEBUG", f"Arguments from AppDaemon config (self.args): {self.args}")
        if 'config' in self.args:
            c = ha_ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]}, ha_args=self.args)
            self.configfile = self.args['config']
            try:
                c.read(self.configfile, encoding='utf-8')
                c.configfile = self.configfile
            except Exception as e:
                hybridlogger.ha_log(logger, self, "ERROR",
                                    f"Error parsing config.ini '{self.configfile}'. No valid configuration file. {e}.")
                c.configfile = None
        else:
            c = ha_ConfigParser(ha_args=self.args)
            c.configfile = None
        set_data_config_path(c)

        self.interval = int(c.get('default', 'interval', 360))
        self.datalogger = DataLogger(c, hass_api=self)

        # running repeatedly
        if self.datalogger.client.use_timer:
            hybridlogger.ha_log(logger, self, "INFO", f"Daemon interval {self.interval} seconds.")
        else:
            hybridlogger.ha_log(logger, self, "INFO", "Daemon starts listening.")
        self.rt = RepeatedJob(c, self.datalogger, hass_api=self)

    def terminate(self):
        hybridlogger.ha_log(logger, self, "INFO", "Stopping Omnikdatalogger...")
        self.rt.stop()
        self.datalogger.terminate()
        hybridlogger.ha_log(logger, self, "INFO", "Omnikdatalogger was stopped")


# Initialization from the commandline
def main(c: ha_ConfigParser, hass_api=None):
    global stopflag
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    # Enabled debugging if the flag is set
    if c.get('default', 'debug', False):
        logger.setLevel(logging.DEBUG)

    set_data_config_path(c)
    datalogger = DataLogger(c, hass_api=hass_api)
    if c.has_option('default', 'interval'):
        # running repeatedly
        if datalogger.client.use_timer:
            hybridlogger.ha_log(logger, hass_api, "INFO", f"Daemon interval {int(c.get('default', 'interval', 360))} seconds.")
        else:
            hybridlogger.ha_log(logger, hass_api, "INFO", "Daemon starts listening.")
        rt = RepeatedJob(c, datalogger, hass_api)
        if hass_api:
            hass_api.rt = rt
        else:
            # loop is for commandline execution only
            try:
                while not stopflag:
                    time.sleep(0.5)
            except KeyboardInterrupt:
                pass
            hybridlogger.ha_log(logger, hass_api, "INFO", "Stopping Omnikdatalogger...")
            rt.stop()
            datalogger.terminate()
            hybridlogger.ha_log(logger, hass_api, "INFO", "Omnikdatalogger was stopped")

    else:
        # running only once
        datalogger.process()
        datalogger.terminate()


def set_data_config_path(config):
    config.data_config_file_path = f"{pathlib.Path(__file__).parent.absolute()}/data_fields.json"
    config.data_config_file_shared = f"{pathlib.Path(__file__).parent.parent.absolute()}" \
                                     "/share/omnikdatalogger/data_fields.json"


def get_yaml_settings(args):
    with open(args.settings, 'r') as stream:
        try:
            settings = yaml.safe_load(stream)
            if args.section in settings:
                index = args.section
            else:
                for key in settings:
                    index = key
                    break
            logging.info("Using section '{1}' from config file '{0}'".
                         format(args.settings, index))
            return settings[index]
        except yaml.YAMLError as exc:
            logging.error("YAML config file '{0}' could not be parsed. Error: {1}".
                          format(args.settings, exc))
            os.sys.exit(1)


def setup_config_parser(args, settings):
    if not settings:
        logging.error("Config files '{0}' has no valid index".
                      format(args.settings))
        os.sys.exit(1)
    elif 'config' in settings:
        try:
            configfile = settings['config']
            c = ha_ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]}, ha_args=settings)
            c.read(configfile, encoding='utf-8')
            c.configfile = configfile
            logging.info("Using config file '{0}' from the 'config' key of YAML file {1}".
                         format(configfile, args.settings))
        except Exception as e:
            logging.error(f"Error parsing config.ini '{configfile}'. No valid configuration file. {e}.")
            os.sys.exit(1)
    else:
        try:
            c = ha_ConfigParser(ha_args=settings)
            c.configfile = None
        except Exception as e:
            logging.error("Config parser problem '{0}'".
                          format(e))
            os.sys.exit(1)
    return c


def get_config_from_files(args):
    c = None
    if os.path.isfile(args.settings):
        if os.path.isfile(args.config):
            logging.warning("Multiple config files found at '{0}' or '{1}', "
                            "using '{0}', ignoring config from command line option '{1}'.'".
                            format(args.settings, args.config))

        settings = get_yaml_settings(args)
        c = setup_config_parser(args, settings)
    if os.path.isfile(args.config) and not os.path.isfile(args.settings):
        ha_args['config'] = args.config
        if args.debug:
            ha_args['debug'] = args.debug
        if args.interval:
            ha_args['interval'] = args.interval
        if args.data_config:
            ha_args['data_config'] = args.data_config
        if args.persistant_cache_file:
            ha_args['persistant_cache_file'] = args.persistant_cache_file
        c = ha_ConfigParser(converters={'list': lambda x: [i.strip() for i in x.split(',')]}, ha_args=ha_args)
        c.read([args.config], encoding='utf-8')
        c.configfile = args.config
    if not c:
        logging.error("No valid configuration found. Exiting!")
        os.sys.exit(1)
    return c


if __name__ == '__main__':
    stopflag = False
    home = os.path.expanduser('~')
    parser = argparse.ArgumentParser()

    parser.add_argument('--settings', default=os.path.join(home, '.omnik/config.yaml'),
                        help='Path to .yaml configuration file', metavar="FILE")
    parser.add_argument('--section', default=None,
                        help='Section to .yaml configuration file to use. Defaults to the first section found.')
    parser.add_argument('--config', default=os.path.join(home, '.omnik/config.ini'),
                        help='Path to configuration file (ini) (DECREPATED!)', metavar="FILE")
    parser.add_argument('--data_config',
                        help='Path to data_fields.json configuration file', metavar="FILE")
    parser.add_argument('--persistant_cache_file',
                        help='Path to writable cache json file to store last power en total energy', metavar="FILE")
    parser.add_argument('--interval', type=int, help='Execute every n seconds')

    parser.add_argument('-d', '--debug', action='store_true', help='Debug mode')

    args = parser.parse_args()

    if not os.path.isfile(args.settings) and not os.path.isfile(args.config):
        logging.critical("Unable to find config at '{0}' or {1}".
                         format(args.settings, args.config))
        parser.print_help()
        os.sys.exit(1)
    elif os.path.isfile(args.config):
        logging.warning("The use of a classig config.ini is decrepated! "
                        "Use the --settings option. See documentation.")

    main(get_config_from_files(args), hass_api=None)



import os
import re
from setuptools import setup, find_packages


def get_version(package):
    """
    Return package version as listed in `__version__` in `__init__.py`.
    """
    path = os.path.join(os.path.dirname(__file__), package, '__init__.py')
    with open(path, 'rb') as f:
        init_py = f.read().decode('utf-8')
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


with open("README.md", "r", encoding="UTF-8") as fh:
    long_description = fh.read()

install_requires = [
    'configparser>=3.7.4',
    'requests>=2.21.0',
    'cachetools>=3.1.1',
    'pytz>=2019.1',
    'paho-mqtt>=1.5.0',
    'dsmr-parser>=0.21',
    'astral>=1.10.1',
    'pyyaml>=5.1',
    'omnikdataloggerproxy>=1.2.0'
]

setup(
    name="omnikdatalogger",
    version=get_version('apps/omnikdatalogger/omnik'),
    license="gpl-3.0",
    author="Jan Bouwhuis",
    author_email="jan@jbsoft.nl",
    description="Omnik Data Logger",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jbouwh/omnikdatalogger",
    package_dir={"": "apps/omnikdatalogger"},
    packages=find_packages(where="apps/omnikdatalogger"),
    data_files=[('share/omnikdatalogger', [
        'apps/omnikdatalogger/data_fields.json',
        'scripts/systemd/omnikdatalogger.service',
        ])
                ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Home Automation',
        'Programming Language :: Python :: 3.7',
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent"
    ],
    install_requires=install_requires,
    scripts=['apps/omnikdatalogger/omniklogger.py'],
)



from omnik.ha_logger import hybridlogger
from omnik.plugin_localproxy import LocalProxyPlugin
import binascii


class HASSAPI(LocalProxyPlugin):

    '''
    This plugin enables you to subscribe to the logger requests using the AppDaemon HASSAPI integration
    Use scripts/proxy/omnikloggerproxy.py to capture the inverter events and publish the data to MQTT (Home Assistant)

    This class makes use the following localproxy client objects
    self.client.msg
    self.client.msgevent
    self.client.semaphore
    self.client.inverters
    self.client.plant_id_list
    '''

    def __init__(self):
        super().__init__()
        hybridlogger.ha_log(self.logger, self.hass_api, "INFO", "localproxy client plugin: HASSAPI")
        self.hass_handle = None
        if not self.hass_api:
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "ERROR", "No HassAPI detected. Use AppDaemon with Home Assistent for this plugin")
            return
        self.logger_entity = self.config.get('client.localproxy.hassapi', 'logger_entity', 'binary_sensor.datalogger')

    def terminate(self):
        if self.hass_handle:
            self.hass_api.cancel_listen_state(self.hass_handle)

    def listen(self):
        if self.hass_api:
            self.hass_handle = self.hass_api.listen_state(self._run, self.logger_entity, attribute='data')
            hybridlogger.ha_log(self.logger, self.hass_api,
                                "INFO", f"HASSapi listening to. '{self.logger_entity}', attribute: 'data'")

    def _run(self, entity, attribute, old, new, kwargs):
        # HASSAPI callback handler
        hybridlogger.ha_log(self.logger, self.hass_api,
                            "INFO",
                            f"HASSapi state change for {self.hass_api.get_state(self.logger_entity, 'inverter', 'n/a')} "
                            f"at {self.hass_api.get_state(self.logger_entity, 'last_update', 'n/a')}")
        # Try to parse payload as json
        try:
            data = binascii.a2b_base64(new)
            if len(data) == 128:
                # Fill data structure
                self.client.semaphore.acquire()
                self.client.msg['data'] = data
                self.client.msg['isSet'] = True
                self.client.msg['plugin'] = __name__
                self.client.semaphore.release()
                # Trigger processing the message
                self.client.msgevent.set()
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "WARNING", f"HASSapi: invalid data received: {new}. Error {e}")



import json
import time
from omnik.ha_logger import hybridlogger

import paho.mqtt.client as mqttclient
import uuid

from omnik.plugin_output import Plugin
import threading


class mqtt(Plugin):

    def __init__(self):
        super().__init__()
        self.name = 'mqtt'
        self.description = 'Write output to internal mqtt facility of Home Assistant'
        # tz = self.config.get('default', 'timezone',
        #                      fallback='Europe/Amsterdam')
        # self.timezone = pytz.timezone(tz)

        self.mqtt_client_name = self.config.get('output.mqtt', 'client_name_prefix',
                                                fallback='ha-mqtt-omniklogger') + "_" + uuid.uuid4().hex
        self.mqtt_host = self.config.get('output.mqtt', 'host', fallback='localhost')
        self.mqtt_port = int(self.config.get('output.mqtt', 'port', fallback='1833'))
        self.mqtt_retain = self.config.getboolean('output.mqtt', 'retain', fallback=False)
        self.mqtt_config_published = {}
        if not self.config.has_option('output.mqtt', 'username') or not self.config.has_option('output.mqtt', 'password'):
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR",
                                "Please specify MQTT username and password in the configuration")
            self.mqtt_username = None
            self.mqtt_password = None
        else:
            self.mqtt_username = self.config.get('output.mqtt', 'username')
            self.mqtt_password = self.config.get('output.mqtt', 'password')

        # mqtt setup
        self.mqtt_client = mqttclient.Client(self.mqtt_client_name)
        self.mqtt_client.on_connect = self._mqtt_on_connect  # bind call back function
        self.mqtt_client.on_disconnect = self._mqtt_on_disconnect  # bind call back function
        self.mqtt_client.logger = self.logger
        self.mqtt_client.hass_api = self.hass_api
        # self.mqtt_client.on_message=mqtt_on_message (not used)
        self.mqtt_client.username_pw_set(self.mqtt_username, self.mqtt_password)
        self.mqtt_client.connect(self.mqtt_host, self.mqtt_port)
        # start processing messages
        self.mqtt_client.loop_start()
        if self.config.has_option('output.mqtt', 'discovery_prefix'):
            self.discovery_prefix = self.config.get('output.mqtt', 'discovery_prefix')
        else:
            self.discovery_prefix = "homeassistant"
        # The mqtt name of the inverter device
        # use the configured device name or the device name from the omnik portal
        # if self.config.has_option('output.mqtt', 'device_name'):
        #    self.device_name = self.config.get('output.mqtt', 'device_name')
        # else:
        #    self.device_name = None
        # self.config.mqtt_fields['<field_name>']['name'] can have an override
        # self.mqtt_field_name_override_init = {}
        # Init topics dict
        self.topics = {}
        # Init config dict
        self.config_pl = {}
        # Make instance to run exclusively
        self.access = threading.Condition(threading.Lock())

    def _init_config(self, msg):
        # Check if init is needed
        asset_classes = set()
        for field in msg:
            if field in self.config.data_field_config:
                asset_class = self.config.data_field_config[field]['asset']
                if asset_class not in asset_classes:
                    asset_classes.add(asset_class)
        # if msg['plant_id'] in self.mqtt_field_name_override_init:
        #    return
        # Init from using mqtt field config (loaded from json)
        # self.config.data_field_config
        # field: name, dev_cla, ic, unit, measurement, filter, asset

        # Assemble topics
        self.topics[msg['plant_id']] = self._topics(msg, asset_classes)

        # Assemble config
        self.config_pl[msg['plant_id']] = self._config_payload(msg, self.topics[msg['plant_id']], asset_classes)

        # Flag init as done
        # self.mqtt_field_name_override_init[msg['plant_id']] = True
        return asset_classes

    def _mqtt_on_connect(self, client, userdata, flags, rc=0, properties=None):
        if rc == 0:
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO", "MQTT connected")
            # subscribe listening (not used)

    def _mqtt_on_disconnect(self, client, userdata, flags, rc=0):
        if rc == 0:
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO", "MQTT disconnected")

    def _topics(self, msg, asset_classes):
        # Init from using mqtt field config (loaded from json)
        # self.config.data_field_config
        # field: name, dev_cla, ic, unit, filter

        topics = {}
        for asset_class in asset_classes:
            topics[asset_class] = {}
            topics[asset_class]['main'] = \
                f"{self.discovery_prefix}/sensor/{self.config.attributes['devicename'][asset_class]}_{msg['plant_id']}"
            topics[asset_class]['state'] = f"{topics[asset_class]['main']}/state"
            topics[asset_class]['attr'] = f"{topics[asset_class]['main']}/attr"
            topics[asset_class]['config'] = {}
            for field in self.config.data_field_config:
                if field in msg:
                    topics[asset_class]['config'][field] = f"{topics[asset_class]['main']}/{field}/config"
        return topics

    def _device_payload(self, msg, asset_classes):
        device_pl = {}
        for asset_class in asset_classes:
            device_pl[asset_class] = {}
            # Set device_name
            device_name = self.config.attributes['devicename'][asset_class]
            # Determine appendixes
            name_appendix = ""
            identifier = self.config.attributes['identifier'][asset_class]
            if identifier in msg and self.config.getboolean('output.mqtt', 'append_plant_id', False):
                name_appendix = f" {msg[identifier]}"
            id_prefix = ""
            if identifier in msg:
                id_prefix = f"{msg[identifier]}_"
            # Device payload
            device_pl[asset_class] = {
                "identifiers": [f"{id_prefix}{asset_class}"],
                "name": f"{device_name}{name_appendix}",
                "mdl": self.config.attributes['model'][asset_class],
                "mf": self.config.attributes['mf'][asset_class]
                }
        return device_pl

    def _config_payload(self, msg, topics, asset_classes):
        # Get device payload
        device_pl = self._device_payload(msg, asset_classes)

        # Fill config_pl dict
        # Init from using mqtt field config (loaded from json)
        # self.config.data_field_config
        config_pl = {}

        # Generate MQTT config topics
        for field in self.config.data_field_config:
            # Only generate config for available data
            if field not in msg:
                # skip publishing non exitent values
                continue
            asset_class = self.config.data_field_config[field]['asset']
            if not asset_class:
                # skip publication
                continue
            # field: name, dev_cla, ic, unit, measurent, filter, asset
            if self.config.has_option('output.mqtt', f'{field}_name'):
                fieldname = self.config.get('output.mqtt', f'{field}_name')
            else:
                fieldname = self.config.data_field_config[field]['name']
            identifier = self.config.attributes['identifier'][asset_class]
            config_pl[field] = {
                "~": f"{topics[asset_class]['main']}",
                "uniq_id": f"{msg[identifier]}_{field}",
                "name": f"{fieldname}",
                "stat_t": "~/state",
                "json_attr_t": "~/attr",
                "val_tpl": f"{{{{(value_json.{field}{self.config.data_field_config[field]['filter']})}}}}",
                "dev": device_pl[asset_class]
                }
            # Set device class if configured
            if self.config.data_field_config[field]['dev_cla']:
                config_pl[field]['dev_cla'] = self.config.data_field_config[field]['dev_cla']
            # Set icon if configured
            if self.config.data_field_config[field]['ic']:
                config_pl[field]['ic'] = f"mdi:{self.config.data_field_config[field]['ic']}"
            # Set unit of measurement if configured
            if self.config.data_field_config[field]['unit']:
                config_pl[field]['unit_of_meas'] = self.config.data_field_config[field]['unit']

        return config_pl

    def _value_payload(self, msg):
        value_pl = {}
        # Generate MQTT config topics
        for field in self.config.data_field_config:
            # field: name, dev_cla, ic, unit, filter
            # Only generate payload for available data
            asset_class = self.config.data_field_config[field]['asset']
            if not asset_class:
                # skip publication
                continue
            if asset_class not in value_pl:
                value_pl[asset_class] = {}
            if field in msg:
                value_pl[asset_class][field] = self.jsonval(msg[field])

        return value_pl

    def _attribute_payload(self, msg, asset_classes):
        attr_pl = {}
        for asset_class in asset_classes:
            apl = self._attribute_payload_asset_class(msg, asset_class)
            if apl:
                attr_pl[asset_class] = apl
        return attr_pl

        # attr_pl[asset_class] = {
        # "inverter": msg['inverter'],
        # "plant_id": msg['plant_id'],
        # "last_update": time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(msg['last_update']))
        # }

    def _attribute_payload_asset_class(self, msg, asset_class):
        attr_pl_class = {}
        for attr in self.config.attributes['asset'][asset_class]:
            attr_pl = self._encode_attribute(msg, attr)
            if attr_pl:
                attr_pl_class[attr] = self._encode_attribute(msg, attr)
        return attr_pl_class

    def _encode_attribute(self, msg, attr):
        if attr not in msg:
            return None
        if attr in self.config.data_field_config:
            return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(msg[attr])) \
                if self.config.data_field_config[attr]['dev_cla'] == 'timestamp' \
                else self.jsonval(msg[attr])
        else:
            return self.jsonval(msg[attr])

    def _publish_config(self, msg):
        new_asset_classes = set()
        if msg['plant_id'] not in self.mqtt_config_published:
            # init mqtt_config_published flag cache
            self.mqtt_config_published[msg['plant_id']] = set()

        for entity in self.config_pl[msg['plant_id']]:
            asset_class = self.config.data_field_config[entity]['asset']
            if asset_class in self.mqtt_config_published[msg['plant_id']] and self.mqtt_retain:
                continue
            else:
                self._publish_config_entity(self.topics[msg['plant_id']][asset_class], self.config_pl[msg['plant_id']], entity)
                new_asset_classes.add(asset_class)
        for item in new_asset_classes:
            self.mqtt_config_published[msg['plant_id']].add(item)

    def _publish_config_entity(self, topics, config_pl, entity):
        try:
            # publish config
            # asset_class = self.config.data_field_config[entity]['asset']
            if self.mqtt_client.publish(topics['config'][entity], json.dumps(config_pl[entity]), retain=self.mqtt_retain):
                hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG",
                                    f"Publishing config {json.dumps(config_pl[entity])} "
                                    f"to {topics['config'][entity]} successful.")
            else:
                hybridlogger.ha_log(self.logger, self.hass_api, "WARNING",
                                    f"Publishing config {json.dumps(config_pl[entity])} "
                                    f"to {topics['config'][entity]} failed!")
        except Exception as e:
            hybridlogger.ha_log(self.logger, self.hass_api, "ERROR",
                                f"Unhandled error publishing config for entity {entity}: {e}")

    def _publish_attributes(self, msg, asset_classes):

        attr_pl = self._attribute_payload(msg, asset_classes)
        for asset_class in asset_classes:
            try:
                # publish attributes
                if self.mqtt_client.publish(self.topics[msg['plant_id']][asset_class]['attr'],
                                            json.dumps(attr_pl[asset_class]), retain=self.mqtt_retain):
                    hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG",
                                        f"Publishing attributes {json.dumps(attr_pl[asset_class])} to "
                                        f"{self.topics[msg['plant_id']][asset_class]['attr']} successful.")
                else:
                    hybridlogger.ha_log(self.logger, self.hass_api, "WARNING",
                                        f"Publishing attributes {json.dumps(attr_pl[asset_class])} to "
                                        f"{self.topics[msg['plant_id']][asset_class]['attr']} failed!")
            except Exception as e:
                hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", f"Unhandled error publishing attributes: {e}")

    def _publish_state(self, topics, value_pl, asset_classes):
        for asset_class in asset_classes:
            try:
                # publish state
                if self.mqtt_client.publish(topics[asset_class]['state'],
                                            json.dumps(value_pl[asset_class]),
                                            retain=self.mqtt_retain):
                    hybridlogger.ha_log(self.logger, self.hass_api, "DEBUG",
                                        f"Publishing state {json.dumps(value_pl[asset_class])} to "
                                        f"{topics[asset_class]['state']} successful.")
                else:
                    hybridlogger.ha_log(self.logger, self.hass_api, "WARNING",
                                        f"Publishing state {json.dumps(value_pl[asset_class])} to "
                                        f"{topics[asset_class]['state']} failed!")
            except Exception as e:
                hybridlogger.ha_log(self.logger, self.hass_api, "ERROR", f"Unhandled error publishing states: {e}")

    def process(self, **args):
        """
        Send data to over mqtt (compliant with Home Assistant MQTT discovery standard
        See: https://www.home-assistant.io/docs/mqtt/discovery/)
        """

        self.access.acquire()

        # Get argument
        msg = args['msg']

        # Set report time in local timezone
        # msg['reporttime'] = time.localtime(msg['last_update'])

        # Assemble config
        asset_classes = self._init_config(msg)

        # Publish config
        self._publish_config(msg)

        # publish attributes
        self._publish_attributes(msg, asset_classes)

        # publish state
        value_pl = self._value_payload(msg)
        self._publish_state(self.topics[msg['plant_id']], value_pl, asset_classes)

        self.access.release()
